#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <errno.h>
#define SEND_BUFFER_SIZE 256000 

int main(int argc, char **argv) {
   int fd; /* file descriptor */
   int nfd; /* inet file descriptor */
   int read_val; /* bytes read */
   int fread_val;
   size_t filelen;
   char *host; /* network host */
   char op; /* read or write operation */
   char *file; /* filename to write or read from  */
   char rfile[200]; 
   char pin[20];
   char buff[257];
   int *bin;
   int send_buffer_size = SEND_BUFFER_SIZE;
   _Bool pinaccepted = 0;
   
   /* modules */
   int open_file(char *path,char op);
   int ipv4_tcp_connect(char *);
   int write_file(int,int *,int);
   int read_file(int,int *,int);

   if(argc == 4) {
      fprintf(stdout,"PIN?(max 10): ");
      fgets(pin,11,stdin); /* read pin number */
      if( (feof(stdin) || ferror(stdin)) ) { /* error handling */
         clearerr(stdin);
         fprintf(stderr,"error while reading from console\n");
         exit(EXIT_FAILURE);
      }
      if(pin[(strlen(pin))-1] == '\n') {
         pin[(strlen(pin))-1] = 0; 
      }
      host = argv[1]; /* node */
      op = argv[2][0]; /* <s> or <r> */
      buff[0]=0; /* strcat(see below) makes a concat on null byte */
      file = argv[3]; /* filepath */

      if(op == 'r') {
         rfile[0]='\0';
         strncat(rfile,file,(sizeof(rfile))-4);
         if(file[(strlen(file))-1] != '/') {
            strcat(rfile,"/");
         }
      }
      if(op == 's') { /* sender operation */
         strcat(buff,"SP:");
         strcat(buff,pin);
         pin[0]=0;
         strcat(pin,buff);
         nfd = ipv4_tcp_connect(host);

         while((setsockopt(nfd,SOL_SOCKET,SO_SNDBUF,&send_buffer_size,sizeof(send_buffer_size))) == -1 ) {
            
            if(errno == ENOBUFS) {
               send_buffer_size -= 100;
            }
            else {
               perror("on setsockopt():");
	       close(nfd);
	       exit(EXIT_FAILURE);
            }
         }
         printf("SEND BUFFER SIZE = %i\n",send_buffer_size);

         if(!(bin = (int *) calloc(send_buffer_size,sizeof(int)))) {
            perror("on calloc()");
            exit(EXIT_FAILURE);
         }

         write(nfd,buff,(strlen(buff)));
         read_val=1;
         while(read_val > 0) {

             read_val = read(nfd,buff,255);

             if(read_val > 0 ) {
                buff[read_val]=0;
                if(!(strcmp(buff,pin))) {
                   printf("PIN Accepted: %s\nWaiting for receiver\n",buff);
                }
                else if(!(strcmp(buff,"Ready"))) {
                   filelen = strlen(file);
                   --filelen;
                   while(file[filelen] == '/') {
                      file[filelen] = 0;
                      --filelen;
                   }
                   while(filelen > 0) {
                      if(file[filelen] == '/') {
                         ++filelen;
                         break;
                      }
                      --filelen;
                   }
                   printf("Ready to send\n");
                   write(nfd,"SET",3);
                   read(nfd,buff,2);
                   if(buff[0] == 'O' && buff[1] == 'K') {
		      if((size_t)(write(nfd,(char *)&file[filelen],
 			             (strlen((char *)&file[filelen])))) != (strlen((char *)&file[filelen]))) 
                      {
        		printf("on write()\n");
                      }
                      else {
                         read(nfd,buff,5);
                         if(buff[0] == 'F' && buff[1] == 'S' && buff[2] == 'E' && buff[3] == 'N' && buff[4] == 'T') {
                            /* last */
                            if( (fd = open_file(file,op)) == -1) {
                                free(bin);
			        close(nfd);
                                exit(EXIT_FAILURE);
                            }
                            else {
                               while((fread_val = read_file(fd,bin,send_buffer_size)) != -1 && fread_val != 0) {
                                  if((write(nfd,bin,fread_val)) == -1) {
                                     perror("on write()");
                                     close(fd);
                                     free(bin);
				     exit(EXIT_FAILURE);
                                  }
                               } 
                               if(fread_val == -1) {
                                  close(nfd);
                               }
                               else {
                                  fprintf(stdout,"End of File\n");
                                  close(fd);
                                  close(nfd);
                               }
                               break;
                            }
                         }
                      }
                   }
                }
                else {
                   printf("PIN Error: %s\n",buff);
                   close(nfd);
                   break;
                }
             }

             else {
                if(read_val == 0) {
                   fprintf(stderr,"Connection Closed\n");
                   close(nfd);
                }
                else {
                   perror("in read()");
                }
             }

         }
         free(bin);
      }
      else if(op == 'r') {
         strcat(buff,"RP:");
         strcat(buff,pin);
         pin[0]=0;
         strcat(pin,buff);
         nfd = ipv4_tcp_connect(host);
         if(!(bin = (int *) calloc(send_buffer_size,sizeof(int)))) {
            perror("on calloc()");
            close(nfd);
            exit(EXIT_FAILURE);
         }
         write(nfd,buff,(strlen(buff)));
         read_val=1;
         while(read_val > 0) {
             read_val = read(nfd,buff,255);
             if(read_val > 0 ) {
                buff[read_val]=0;
                if(!(strcmp(buff,pin))) {
                   printf("PIN Accepted: %s\nWaiting for sender\n",buff);
		   pinaccepted=1;
                   write(nfd,"OK",2);
                }
		else if(pinaccepted) {
		   /* last start storing */
                   strcat(rfile,buff);
                   printf("Writing to: %s\n",rfile);
                   if( (fd = open_file(rfile,op)) == -1) {
                      close(nfd);
                      free(bin);
                      exit(EXIT_FAILURE);
                   } 
                   while((read_val = read(nfd,bin,send_buffer_size)) != -1 && read_val != 0) {
                      if((fread_val = write_file(fd,bin,read_val)) == -1) {
                         close(nfd);
                         free(bin);
                         exit(EXIT_FAILURE);
                      }
                   }
                   if(read_val == 0) {
                      printf("Connection Closed\n");
		      close(fd);
                      close(nfd);
                   }
                   else {
                      perror("on read()");
                      close(fd);
                   }
                   break;
                }
                else {
                   printf("PIN Error: %s\n",buff);
                   close(nfd);
                   break;
                }
             }

             else {
                if(read_val == 0) {
                   fprintf(stderr,"Connection Closed\n");
                   close(nfd);
                }
                else if(read_val == -1) {
                   perror("in read()");
                }
             }

         }
         free(bin);
      }
      
   }

   else {
      fprintf(stderr,"Desc:\nSend or Receive a file using a PIN Number\n\nUsage:\nSend a file through host\n\t%s <host> s <file>\n\nReceive a file to be stored in file directory through host\n\t%s <host> r <file directory>\n",argv[0],argv[0]);
   }

   return 0;
}
